# --------------------------------------------------------------------------------
# Author: Bassel Mokabel
# LinkedIn: https://www.linkedin.com/in/bassel-mokabel/
# --------------------------------------------------------------------------------
# This file has been created by Bassel Mokabel.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------------------

import openai
import json
import time
import pandas as pd

# Setup parameters
openai.api_key = "sk-your-api-key"  # Replace with your OpenAI API key
gpt_model = 'gpt-4-1106-preview'  # The model to be used - at this time you may have newer models to choose from
completion = ''   # Stores openAI completion response
openai_intent = ''   # Predited Intent by openAI
openai_prompt_tokens = '' # Stores the number of prompt Token
prompt_token_cost = 0.01   # The cost of 1000 prompt tokens 1 cent = $0.01 Refer to OpenAI Pricing for the gpt_model mentioned above.
openai_completion_tokens = '' # stores the number of completion token (openAI response)
completion_token_cost = 0.03   # The cost of 1000 completion tokens  3 cent = $0.03
openai_total_tokens = '' # Stores the total number of tokens (prompt + completion above)

# System Content  
SystemContent = """"
Role: 
You are an Intent Classifier for Virtual Assistant Conversations.

Context:
You are analyzing conversations between a virtual assistant from a Telco Company and customer who requested a repair appointment.
Amongst other things, the assistant helps the customer to confirm, reschedule, cancel, or inquire about the appointment and the technician.

Task: 
Determine the user's intent based on their input during interactions with a virtual assistant.
Your role is to classify each user response according to the predefined intent list below.
Predefined Intent List:
1. call request: The user requests to have an agent call or contact them and provides a phone number.
2. tech call before arrival: The user requests to have a call, usually asking the tech to call prior to the appointment or if the tech is late.
3. cancellation: The user requests to cancel their scheduled service or appointment and does not state it was fixed.
4. confirmation: The user confirms their appointment, indicating agreement or approval to maintain the scheduled time.
5. issue fixed: The user state that their telecommunication issue has been resolved or is no longer present.
6. issue not fixed: The user reports that their telecommunication issue remains unresolved.
7. reschedule: The user request to reschedule or change the date and/or time of their appointment.
8. already rescheduled: The user states that they have already rescheduled or change or altered the appointment date and/or time.
9. different time slot: The user requests to have the appointment on the same date it was scheduled but on a different time slot.
10. where is the tech: The user seeks information regarding the technician's estimated time of arrival, questions the technician's attendance, or reports that the technician has not yet arrived, is delayed, or has failed to appear as scheduled.
11. wrong person: The user states that they are not the intended recipient of the communication or the assistant has reached the wrong person or wrong number.
12. contact details provided: The user provides contact information, such as a phone number, email, or address. No request for a phone call is made; the details are simply shared.
13. wrong time: The user states that there is a discrepancy in the communicated appointment time or date, or mentions being informed previously of a different time and date of the appointment.
14. stop communications: The user requests to stop all communications or expresses a desire to no longer be contacted. If stop exist in use request, the intent should be stop.
15. out of scope: The user makes a statement that does not relate to the specified list of intents or falls outside the assistant scope and not part of the above intents.

Output:
Examples of output format:
{"intent":"call request"}
"""


def create_output_files(csv_file_name,excel_file_name):
    # The columns for the CSV file
    columns = [
        "assistant", "user_response", "actual_intent", "openai_intent",
        "classification", "openai_completion_tokens", "openai_completion_cost",
        "openai_prompt_tokens", "openai_prompt_cost", "openai_total_tokens", "openai_total_cost"
    ]

    # Create an empty DataFrame with these columns
    df = pd.DataFrame(columns=columns)

    # Save the DataFrame to a CSV file    
    df.to_csv(csv_file_name, index=False)

    # Save the DataFrame to an Excel file
    df.to_excel(excel_file_name, index=False)

    return True

def read_source_file(file_name, sample):
    # Read the Excel file
    conversations = pd.read_excel(file_name)
    # source file columns are: assistant,user_response
    # assistant > last assistant response
    # user_response > last user response to assistant response
    # actual_intent	> actual intent / label
    # openai_intent > predicted intent by openAI
    # classification > actual vs opean_ai === correct 1 classification or wrong 0
    # openai_completion_tokens, openai_completion_cost, openai_prompt_tokens, openai_prompt_cost, openai_total_tokens, openai_total_cost

    # Lowercase the 'actual_intent' column for case-insensitive comparison
    conversations['actual_intent'] = conversations['actual_intent'].str.lower()

    # Identify all unique intents in the filtered DataFrame
    unique_intents = conversations['actual_intent'].unique()

    # Filter and randomly select 100 samples from each intent
    sampled_conversations = []  # A list to hold the sampled DataFrames

    for intent in unique_intents:
        # Filter the DataFrame for the current intent
        conversations_specific_intent = conversations[conversations['actual_intent'] == intent]

        # Randomly select 100 rows if there are enough, otherwise take all available
        sample_conversation = conversations_specific_intent.sample(n=min(sample, len(conversations_specific_intent)), random_state=10)  # random_state for reproducibility
        
        # Add the sampled DataFrame to the list
        sampled_conversations.append(sample_conversation)

    # Concatenate all the sampled DataFrames back together
    conversations_df = pd.concat(sampled_conversations, ignore_index=True)

    return conversations_df


def createMessages(conversation):

    initialAssistant = """
        Hi, It\'s Telco comp, our technician is scheduled to arrive and repair your internet service. Please confirm the appointment time still works. If you need to reschedule, cancel or your problem is fixed and you no longer need a tech, let us know here! Reply STOP to opt out.
    """

    # Create Initial Set of Messages consisting System Content and Initial Assistant Message.
    messages = [
        { "role": "system", "content" : SystemContent},
        { "role": "assistant", "content" : initialAssistant},
        ]
    
    # Extract assistant response from the source file
    assistant = conversation['assistant']    # last virtual assistant message before the last user response that we are running classification for     
    # create assistant message 
    assistant_message = {
            "role": 'assistant',
            "content": assistant
        }
    # append the assitant message to the overall messages
    messages.append(assistant_message)

    # Extract user response from conversation
    user_response = conversation['user_response']
    user_message = {
            "role": 'user',
            "content": str(user_response)
        }

    messages.append(user_message)

    return messages



# Function to call openAI completion
def chatComplete(messages):

    # Due to timeout when using openAI we are waiting 10 seconds between API Calls.
    print('Waiting 10 Seconds before sending a request to OpenAI...')
    time.sleep(10)

    completion = openai.ChatCompletion.create(        
        model= gpt_model,
        max_tokens=2000,
        messages = messages,
        temperature = 1,
        top_p=0.5,
        frequency_penalty=0.5,
        presence_penalty=0
    )
    
    response = completion['choices'][0]['message']['content']    
    usage = completion['usage']

    # return both content/response and the usage    
    return response, usage


def extractAndCompareIntent(response,actual_intent):  
          
    #  Extract intent from openAI response
    if 'intent' in response:
        # Parse the JSON
        data = json.loads(response)
        # Extract the response field
        openai_intent = data['intent'].lower()
    else:
        print('not a JSON object')
        openai_intent = ''

    if ( openai_intent == actual_intent):
        classificationResult = 'correct'  # can be replaced by 1 but will have to update the notebook
    else:
        classificationResult = 'wrong'  # can be replaced by 0 but will have to update the notebook
    
    return classificationResult,openai_intent



def classifyIntents(source_excel_file_name,output_csv_file_name,output_excel_file_name,intentSample):

    # create files
    create_output_files(output_csv_file_name,output_excel_file_name)

    # Read the source data and extract all conversations    
    conversations = read_source_file(source_excel_file_name,intentSample)

    # Iterate over each line / conversation between virtual assistant and the customer
    for index, converation in conversations.iterrows():

        print("*****************",index,"*****************")        
        # Extract the assistant message from the source file
        assistant = converation['assistant']
        # Extract the last user response from the source file
        user_response = converation['user_response']
        # Extract the actual intent from the source file
        actual_intent = converation['actual_intent']  # this was labeled by labeling professionals

        # Create Messages for OpenAI Completion
        messages = createMessages(converation)

        #  Send requests to OpenAI API and get response and usage   
        response, usage = chatComplete(messages)       

        # Compare the classified Intent with the actual_intent and return the results and the predicted intent
        intentResult, openai_intent = extractAndCompareIntent(response,actual_intent)        

        # update the predicted intent column - openai_intent
        conversations.at[index, 'openai_intent'] = openai_intent

        # Number of prompt tokens sent back by openAI
        openai_prompt_tokens = usage['prompt_tokens']
        conversations.at[index, 'openai_prompt_tokens'] = openai_prompt_tokens
        # Cost of the prompt tokens
        openai_prompt_cost = int(usage['prompt_tokens'])*prompt_token_cost/1000
        conversations.at[index, 'openai_prompt_cost'] = openai_prompt_cost
        
        # Number of completion tokens sent back by openAI
        openai_completion_tokens = usage['completion_tokens'] 
        conversations.at[index, 'openai_completion_tokens'] = openai_completion_tokens
        # Cost of the completion tokens
        openai_completion_cost = int(usage['completion_tokens'])*completion_token_cost/1000
        conversations.at[index, 'openai_completion_cost'] = openai_completion_cost

        # The total number of tokens sent back by openAI
        openai_total_tokens = usage['total_tokens']
        conversations.at[index, 'openai_total_tokens'] = openai_total_tokens

        # The total cost of tokens
        openai_total_cost = openai_prompt_cost + openai_completion_cost
        conversations.at[index, 'openai_total_cost'] = openai_total_cost

        # Add the Classification Result to existing DataFrame
        conversations.at[index, 'classification'] = intentResult

        print(f'User Response {user_response}',f'Actual Intent: {actual_intent}',f'Predicted Intent: {openai_intent}',intentResult,sep="\n")        

        # Create Dictionary 
        # This will be used to update the csv while the classification is happening
        # Due to timeout, we will update a csv after each row
        data_to_append = {'assistant': assistant, 
                        'user_response': user_response, 
                        'actual_intent': actual_intent,
                        'openai_intent':openai_intent,
                        'classification': intentResult,
                        'openai_completion_tokens': openai_completion_tokens,
                        'openai_completion_cost': str(float("{:.8f}".format(openai_completion_cost))),
                        'openai_prompt_tokens': openai_prompt_tokens,
                        'openai_prompt_cost': str(float("{:.8f}".format(openai_prompt_cost))),
                        'openai_total_tokens': openai_total_tokens,
                        'openai_total_cost': str(float("{:.8f}".format(openai_total_cost)))
                    }
        # Create a DataFrame from the data you want to append
        df_to_append = pd.DataFrame(data_to_append, index=[0])
        # append to the csv file after each row
        df_to_append.to_csv(output_csv_file_name, mode='a', header=False, index=False)

    # Save the updated DataFrame to a new Excel file after  all tests have been completed    
    conversations.to_excel(output_excel_file_name, index=False)
    print(f"Updated data has been saved to {output_excel_file_name}")


def main():
    #  Run the function with the given parameters
    intentSample = 20 # get 20 conversations of each intent / label
    source_excel_file_name = 'conversations_source.xlsx'
    output_csv_file_name = 'intent_classification_results.csv'
    output_excel_file_name = 'intent_classification_results.xlsx'
    classifyIntents(source_excel_file_name,output_csv_file_name,output_excel_file_name,intentSample)

# run the main function
main()